@@ -40,6 +40,14 @@ typedef enum {
 	HEADSET_LOCK_WRITE = 1 << 1,
 } headset_lock_t;
 
+typedef void (*headset_state_cb) (struct audio_device *dev,
+					headset_state_t old_state,
+					headset_state_t new_state,
+					void *user_data);
+
+unsigned int headset_add_state_cb(headset_state_cb cb, void *user_data);
+gboolean headset_remove_state_cb(unsigned int id);
+
 typedef void (*headset_stream_cb_t) (struct audio_device *dev, void *user_data);
 
 void headset_connect_cb(GIOChannel *chan, GError *err, gpointer user_data);
@@ -58,15 +66,12 @@ void headset_update(struct audio_device *dev, uint16_t svc,
 
 unsigned int headset_config_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data);
 unsigned int headset_request_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data);
 unsigned int headset_suspend_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data);
 gboolean headset_cancel_stream(struct audio_device *dev, unsigned int id);
 
@@ -88,6 +93,7 @@ gboolean headset_get_sco_hci(struct audio_device *dev);
 
 gboolean headset_is_active(struct audio_device *dev);
 
+headset_lock_t headset_get_lock(struct audio_device *dev);
 gboolean headset_lock(struct audio_device *dev, headset_lock_t lock);
 gboolean headset_unlock(struct audio_device *dev, headset_lock_t lock);
 gboolean headset_suspend(struct audio_device *dev, void *data);
