@@ -36,7 +36,10 @@ void audio_manager_exit(void);
 
 gboolean server_is_enabled(bdaddr_t *src, uint16_t svc);
 
-struct audio_device *manager_find_device(const bdaddr_t *bda, const char *interface,
+struct audio_device *manager_find_device(const char *path,
+					const bdaddr_t *src,
+					const bdaddr_t *dst,
+					const char *interface,
 					gboolean connected);
 
 struct audio_device *manager_get_device(const bdaddr_t *src,
