@@ -50,6 +50,8 @@
 #include "unix.h"
 #include "glib-helper.h"
 
+#define check_nul(str) (str[sizeof(str) - 1] == '\0')
+
 typedef enum {
 	TYPE_NONE,
 	TYPE_HEADSET,
@@ -66,7 +68,6 @@ struct a2dp_data {
 };
 
 struct headset_data {
-	headset_lock_t lock;
 	gboolean locked;
 };
 
@@ -75,12 +76,13 @@ struct unix_client {
 	GSList *caps;
 	service_type_t type;
 	char *interface;
+	uint8_t seid;
 	union {
 		struct a2dp_data a2dp;
 		struct headset_data hs;
 	} d;
 	int sock;
-	int access_mode;
+	int lock;
 	int data_fd; /* To be deleted once two phase configuration is fully implemented */
 	unsigned int req_id;
 	unsigned int cb_id;
@@ -93,27 +95,9 @@ static int unix_sock = -1;
 
 static void client_free(struct unix_client *client)
 {
-	struct a2dp_data *a2dp;
-
-	if (client->req_id)
+	if (client->cancel && client->req_id > 0)
 		client->cancel(client->dev, client->req_id);
 
-	switch (client->type) {
-	case TYPE_SINK:
-	case TYPE_SOURCE:
-		a2dp = &client->d.a2dp;
-		if (client->cb_id > 0)
-			avdtp_stream_remove_cb(a2dp->session, a2dp->stream,
-								client->cb_id);
-		if (a2dp->sep)
-			a2dp_sep_unlock(a2dp->sep, a2dp->session);
-		if (a2dp->session)
-			avdtp_unref(a2dp->session);
-		break;
-	default:
-		break;
-	}
-
 	if (client->sock >= 0)
 		close(client->sock);
 
@@ -230,22 +214,12 @@ static void stream_state_changed(struct avdtp_stream *stream,
 	}
 }
 
-static void headset_discovery_complete(struct audio_device *dev, void *user_data)
+static uint8_t headset_generate_capability(struct audio_device *dev,
+						codec_capabilities_t *codec)
 {
-	struct unix_client *client = user_data;
-	char buf[BT_SUGGESTED_BUFFER_SIZE];
-	struct bt_get_capabilities_rsp *rsp = (void *) buf;
-	codec_capabilities_t *codec;
 	pcm_capabilities_t *pcm;
 
-	client->req_id = 0;
-
-	if (!dev)
-		goto failed;
-
-	memset(buf, 0, sizeof(buf));
-
-	codec = (void *) rsp->data;
+	codec->seid = BT_A2DP_SEID_RANGE + 1;
 	codec->transport = BT_CAPABILITIES_TRANSPORT_SCO;
 	codec->type = BT_HFP_CODEC_PCM;
 	codec->length = sizeof(*pcm);
@@ -256,10 +230,35 @@ static void headset_discovery_complete(struct audio_device *dev, void *user_data
 		pcm->flags |= BT_PCM_FLAG_NREC;
 	if (!headset_get_sco_hci(dev))
 		pcm->flags |= BT_PCM_FLAG_PCM_ROUTING;
+	codec->configured = headset_is_active(dev);
+	codec->lock = headset_get_lock(dev);
+
+	return codec->length;
+}
+
+static void headset_discovery_complete(struct audio_device *dev, void *user_data)
+{
+	struct unix_client *client = user_data;
+	char buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_get_capabilities_rsp *rsp = (void *) buf;
+	uint8_t length;
+
+	client->req_id = 0;
+
+	if (!dev)
+		goto failed;
+
+	memset(buf, 0, sizeof(buf));
+
+	length = headset_generate_capability(dev, (void *) rsp->data);
 
 	rsp->h.type = BT_RESPONSE;
 	rsp->h.name = BT_GET_CAPABILITIES;
-	rsp->h.length = sizeof(*rsp) + codec->length;
+	rsp->h.length = sizeof(*rsp) + length;
+
+	ba2str(&dev->src, rsp->source);
+	ba2str(&dev->dst, rsp->destination);
+	strncpy(rsp->object, dev->path, sizeof(rsp->object));
 
 	unix_ipc_sendmsg(client, &rsp->h);
 
@@ -287,8 +286,6 @@ static void headset_setup_complete(struct audio_device *dev, void *user_data)
 	rsp->h.name = BT_SET_CONFIGURATION;
 	rsp->h.length = sizeof(*rsp);
 
-	rsp->transport = BT_CAPABILITIES_TRANSPORT_SCO;
-	rsp->access_mode = client->access_mode;
 	rsp->link_mtu = 48;
 
 	client->data_fd = headset_get_sco_fd(dev);
@@ -308,25 +305,15 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 	char buf[BT_SUGGESTED_BUFFER_SIZE];
 	struct bt_start_stream_rsp *rsp = (void *) buf;
 	struct bt_new_stream_ind *ind = (void *) buf;
-	struct headset_data *hs = &client->d.hs;
 
 	client->req_id = 0;
 
 	if (!dev)
 		goto failed;
 
-	if (!hs->locked)
-		hs->locked = headset_lock(dev, hs->lock);
-	if (!hs->locked) {
-		error("Unable to lock headset");
-		goto failed;
-	}
-
 	client->data_fd = headset_get_sco_fd(dev);
 	if (client->data_fd < 0) {
 		error("Unable to get a SCO fd");
-		headset_unlock(dev, hs->lock);
-		hs->locked = FALSE;
 		goto failed;
 	}
 
@@ -346,8 +333,6 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 
 	if (unix_sendmsg_fd(client->sock, client->data_fd) < 0) {
 		error("unix_sendmsg_fd: %s(%d)", strerror(errno), errno);
-		headset_unlock(client->dev, hs->lock);
-		hs->locked = FALSE;
 		goto failed;
 	}
 
@@ -382,22 +367,78 @@ failed:
 	client->dev = NULL;
 }
 
+static void print_mpeg12(struct mpeg_codec_cap *mpeg)
+{
+	debug("Media Codec: MPEG12"
+		" Channel Modes: %s%s%s%s"
+		" Frequencies: %s%s%s%s%s%s"
+		" Layers: %s%s%s"
+		" CRC: %s",
+		mpeg->channel_mode & MPEG_CHANNEL_MODE_MONO ? "Mono " : "",
+		mpeg->channel_mode & MPEG_CHANNEL_MODE_DUAL_CHANNEL ?
+		"DualChannel " : "",
+		mpeg->channel_mode & MPEG_CHANNEL_MODE_STEREO ? "Stereo " : "",
+		mpeg->channel_mode & MPEG_CHANNEL_MODE_JOINT_STEREO ?
+		"JointStereo " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_16000 ? "16Khz " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_22050 ? "22.05Khz " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_24000 ? "24Khz " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_32000 ? "32Khz " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_44100 ? "44.1Khz " : "",
+		mpeg->frequency & MPEG_SAMPLING_FREQ_48000 ? "48Khz " : "",
+		mpeg->layer & MPEG_LAYER_MP1 ? "1 " : "",
+		mpeg->layer & MPEG_LAYER_MP2 ? "2 " : "",
+		mpeg->layer & MPEG_LAYER_MP3 ? "3 " : "",
+		mpeg->crc ? "Yes" : "No");
+}
+
+static void print_sbc(struct sbc_codec_cap *sbc)
+{
+	debug("Media Codec: SBC"
+		" Channel Modes: %s%s%s%s"
+		" Frequencies: %s%s%s%s"
+		" Subbands: %s%s"
+		" Blocks: %s%s%s%s"
+		" Bitpool: %d-%d",
+		sbc->channel_mode & SBC_CHANNEL_MODE_MONO ? "Mono " : "",
+		sbc->channel_mode & SBC_CHANNEL_MODE_DUAL_CHANNEL ?
+		"DualChannel " : "",
+		sbc->channel_mode & SBC_CHANNEL_MODE_STEREO ? "Stereo " : "",
+		sbc->channel_mode & SBC_CHANNEL_MODE_JOINT_STEREO ? "JointStereo" : "",
+		sbc->frequency & SBC_SAMPLING_FREQ_16000 ? "16Khz " : "",
+		sbc->frequency & SBC_SAMPLING_FREQ_32000 ? "32Khz " : "",
+		sbc->frequency & SBC_SAMPLING_FREQ_44100 ? "44.1Khz " : "",
+		sbc->frequency & SBC_SAMPLING_FREQ_48000 ? "48Khz " : "",
+		sbc->subbands & SBC_SUBBANDS_4 ? "4 " : "",
+		sbc->subbands & SBC_SUBBANDS_8 ? "8 " : "",
+		sbc->block_length & SBC_BLOCK_LENGTH_4 ? "4 " : "",
+		sbc->block_length & SBC_BLOCK_LENGTH_8 ? "8 " : "",
+		sbc->block_length & SBC_BLOCK_LENGTH_12 ? "12 " : "",
+		sbc->block_length & SBC_BLOCK_LENGTH_16 ? "16 " : "",
+		sbc->min_bitpool, sbc->max_bitpool);
+}
+
 static int a2dp_append_codec(struct bt_get_capabilities_rsp *rsp,
-				struct avdtp_service_capability *cap)
+				struct avdtp_service_capability *cap,
+				uint8_t seid,
+				uint8_t configured,
+				uint8_t lock)
 {
 	struct avdtp_media_codec_capability *codec_cap = (void *) cap->data;
 	codec_capabilities_t *codec = (void *) rsp + rsp->h.length;
-	int space_left = BT_SUGGESTED_BUFFER_SIZE - rsp->h.length;
+	size_t space_left;
 
-	if (space_left <= 0)
+	if (rsp->h.length > BT_SUGGESTED_BUFFER_SIZE)
 		return -ENOMEM;
 
+	space_left = BT_SUGGESTED_BUFFER_SIZE - rsp->h.length;
+
 	/* endianess prevent direct cast */
 	if (codec_cap->media_codec_type == A2DP_CODEC_SBC) {
 		struct sbc_codec_cap *sbc_cap = (void *) codec_cap;
 		sbc_capabilities_t *sbc = (void *) codec;
 
-		if ((space_left - (int) sizeof(sbc_capabilities_t)) < 0)
+		if (space_left < sizeof(sbc_capabilities_t))
 			return -ENOMEM;
 
 		codec->length = sizeof(sbc_capabilities_t);
@@ -409,11 +450,14 @@ static int a2dp_append_codec(struct bt_get_capabilities_rsp *rsp,
 		sbc->block_length = sbc_cap->block_length;
 		sbc->min_bitpool = sbc_cap->min_bitpool;
 		sbc->max_bitpool = sbc_cap->max_bitpool;
+
+		print_sbc(sbc_cap);
+		codec->type = BT_A2DP_SBC_SINK;
 	} else if (codec_cap->media_codec_type == A2DP_CODEC_MPEG12) {
 		struct mpeg_codec_cap *mpeg_cap = (void *) codec_cap;
 		mpeg_capabilities_t *mpeg = (void *) codec;
 
-		if ((space_left - (int) sizeof(mpeg_capabilities_t)) < 0)
+		if (space_left < sizeof(mpeg_capabilities_t))
 			return -ENOMEM;
 
 		codec->length = sizeof(mpeg_capabilities_t);
@@ -424,24 +468,36 @@ static int a2dp_append_codec(struct bt_get_capabilities_rsp *rsp,
 		mpeg->frequency = mpeg_cap->frequency;
 		mpeg->mpf = mpeg_cap->mpf;
 		mpeg->bitrate = mpeg_cap->bitrate;
+
+		print_mpeg12(mpeg_cap);
+		codec->type = BT_A2DP_MPEG12_SINK;
 	} else {
-		int codec_length;
+		size_t codec_length, type_length, total_length;
 
 		codec_length = cap->length - (sizeof(struct avdtp_service_capability)
 				+ sizeof(struct avdtp_media_codec_capability));
+		type_length = sizeof(codec_cap->media_codec_type);
+		total_length = type_length + codec_length +
+				sizeof(codec_capabilities_t);
 
-		if ((space_left - codec_length - (int) sizeof(codec_capabilities_t)) < 0)
+		if (space_left < total_length)
 			return -ENOMEM;
 
-		codec->length = codec_length + sizeof(codec_capabilities_t);
-		memcpy(codec->data, codec_cap->data, codec_length);
+		codec->length = total_length;
+		memcpy(codec->data, &codec_cap->media_codec_type, type_length);
+		memcpy(codec->data + type_length, codec_cap->data,
+			codec_length);
+		codec->type = BT_A2DP_UNKNOWN_SINK;
 	}
 
-	codec->type = codec_cap->media_codec_type;
+	codec->seid = seid;
+	codec->configured = configured;
+	codec->lock = lock;
 	rsp->h.length += codec->length;
 
-	debug("Append codec %d - length %d - total %d", codec->type,
-			codec->length, rsp->h.length);
+	debug("Append %s seid %d - length %d - total %d",
+		configured ? "configured" : "", seid, codec->length,
+		rsp->h.length);
 
 	return 0;
 }
@@ -470,17 +526,56 @@ static void a2dp_discovery_complete(struct avdtp *session, GSList *seps,
 	rsp->h.type = BT_RESPONSE;
 	rsp->h.name = BT_GET_CAPABILITIES;
 	rsp->h.length = sizeof(*rsp);
+	ba2str(&client->dev->src, rsp->source);
+	ba2str(&client->dev->dst, rsp->destination);
+	strncpy(rsp->object, client->dev->path, sizeof(rsp->object));
 
 	for (l = seps; l; l = g_slist_next(l)) {
 		struct avdtp_remote_sep *rsep = l->data;
+		struct a2dp_sep *sep;
 		struct avdtp_service_capability *cap;
+		struct avdtp_stream *stream;
+		uint8_t type, seid, configured = 0, lock = 0;
+		GSList *cl;
+
+		type = avdtp_get_type(rsep);
+
+		if (type != AVDTP_SEP_TYPE_SINK)
+			continue;
 
 		cap = avdtp_get_codec(rsep);
 
 		if (cap->category != AVDTP_MEDIA_CODEC)
 			continue;
 
-		a2dp_append_codec(rsp, cap);
+		seid = avdtp_get_seid(rsep);
+
+		if (client->seid != 0 && client->seid != seid)
+			continue;
+
+		stream = avdtp_get_stream(rsep);
+		if (stream) {
+			configured = 1;
+			if (client->seid == seid)
+				cap = avdtp_stream_get_codec(stream);
+		}
+
+		for (cl = clients; cl; cl = cl->next) {
+			struct unix_client *c = cl->data;
+			struct a2dp_data *ca2dp = &c->d.a2dp;
+
+			if (ca2dp && ca2dp->session == session &&
+					c->seid == seid) {
+				lock = c->lock;
+				break;
+			}
+		}
+
+		sep = a2dp_get_sep(session, stream);
+		if (sep && a2dp_sep_get_lock(sep))
+			lock = BT_WRITE_LOCK;
+
+		a2dp_append_codec(rsp, cap, seid, configured, lock);
 	}
 
 	unix_ipc_sendmsg(client, &rsp->h);
@@ -509,20 +604,16 @@ static void a2dp_config_complete(struct avdtp *session, struct a2dp_sep *sep,
 	uint16_t imtu, omtu;
 	GSList *caps;
 
+	client->req_id = 0;
+
 	if (err)
 		goto failed;
 
 	memset(buf, 0, sizeof(buf));
-	client->req_id = 0;
 
 	if (!stream)
 		goto failed;
 
-	if (!a2dp_sep_lock(sep, session)) {
-		error("Unable to lock A2DP source SEP");
-		goto failed;
-	}
-
 	if (client->cb_id > 0)
 		avdtp_stream_remove_cb(a2dp->session, a2dp->stream,
 								client->cb_id);
@@ -540,9 +631,6 @@ static void a2dp_config_complete(struct avdtp *session, struct a2dp_sep *sep,
 	rsp->h.name = BT_SET_CONFIGURATION;
 	rsp->h.length = sizeof(*rsp);
 
-	rsp->transport = BT_CAPABILITIES_TRANSPORT_A2DP;
-	client->access_mode = BT_CAPABILITIES_ACCESS_MODE_WRITE;
-	rsp->access_mode = client->access_mode;
 	/* FIXME: Use imtu when fd_opt is CFG_FD_OPT_READ */
 	rsp->link_mtu = omtu;
 
@@ -556,10 +644,6 @@ static void a2dp_config_complete(struct avdtp *session, struct a2dp_sep *sep,
 failed:
 	error("config failed");
 
-	if (a2dp->sep) {
-		a2dp_sep_unlock(a2dp->sep, a2dp->session);
-		a2dp->sep = NULL;
-	}
 	unix_ipc_error(client, BT_SET_CONFIGURATION, EIO);
 
 	avdtp_unref(a2dp->session);
@@ -604,10 +688,6 @@ static void a2dp_resume_complete(struct avdtp *session,
 failed:
 	error("resume failed");
 
-	if (a2dp->sep) {
-		a2dp_sep_unlock(a2dp->sep, a2dp->session);
-		a2dp->sep = NULL;
-	}
 	unix_ipc_error(client, BT_START_STREAM, EIO);
 
 	if (client->cb_id > 0) {
@@ -645,10 +725,6 @@ static void a2dp_suspend_complete(struct avdtp *session,
 failed:
 	error("suspend failed");
 
-	if (a2dp->sep) {
-		a2dp_sep_unlock(a2dp->sep, a2dp->session);
-		a2dp->sep = NULL;
-	}
 	unix_ipc_error(client, BT_STOP_STREAM, EIO);
 
 	avdtp_unref(a2dp->session);
@@ -697,6 +773,100 @@ failed:
 	unix_ipc_error(client, BT_GET_CAPABILITIES, err ? : EIO);
 }
 
+static void open_complete(struct audio_device *dev, void *user_data)
+{
+	struct unix_client *client = user_data;
+	char buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_open_rsp *rsp = (void *) buf;
+
+	memset(buf, 0, sizeof(buf));
+
+	rsp->h.type = BT_RESPONSE;
+	rsp->h.name = BT_OPEN;
+	rsp->h.length = sizeof(*rsp);
+
+	ba2str(&dev->src, rsp->source);
+	ba2str(&dev->dst, rsp->destination);
+	strncpy(rsp->object, dev->path, sizeof(rsp->object));
+
+	unix_ipc_sendmsg(client, &rsp->h);
+
+	return;
+}
+
+static void start_open(struct audio_device *dev, struct unix_client *client)
+{
+	struct a2dp_data *a2dp;
+	struct headset_data *hs;
+	struct avdtp_remote_sep *rsep;
+
+	switch (client->type) {
+	case TYPE_SINK:
+		a2dp = &client->d.a2dp;
+
+		if (!a2dp->session)
+			a2dp->session = avdtp_get(&dev->src, &dev->dst);
+
+		if (!a2dp->session) {
+			error("Unable to get a session");
+			goto failed;
+		}
+
+		if (a2dp->sep) {
+			error("Client already has an opened session");
+			goto failed;
+		}
+
+		rsep = avdtp_get_remote_sep(a2dp->session, client->seid);
+		if (!rsep) {
+			error("Invalid seid %d", client->seid);
+			goto failed;
+		}
+
+		a2dp->sep = a2dp_source_get(a2dp->session, rsep);
+		if (!a2dp->sep) {
+			error("seid %d not available or locked", client->seid);
+			goto failed;
+		}
+
+		if (!a2dp_sep_lock(a2dp->sep, a2dp->session)) {
+			error("Unable to open seid %d", client->seid);
+			a2dp->sep = NULL;
+			goto failed;
+		}
+
+		break;
+
+	case TYPE_HEADSET:
+		hs = &client->d.hs;
+
+		if (hs->locked) {
+			error("Client already has an opened session");
+			goto failed;
+		}
+
+		hs->locked = headset_lock(dev, client->lock);
+		if (!hs->locked) {
+			error("Unable to open seid %d", client->seid);
+			goto failed;
+		}
+		break;
+
+	default:
+		error("No known services for device");
+		goto failed;
+	}
+
+	client->dev = dev;
+
+	open_complete(dev, client);
+
+	return;
+
+failed:
+	unix_ipc_error(client, BT_OPEN, EINVAL);
+}
+
 static void start_config(struct audio_device *dev, struct unix_client *client)
 {
 	struct a2dp_data *a2dp;
@@ -715,31 +885,27 @@ static void start_config(struct audio_device *dev, struct unix_client *client)
 			goto failed;
 		}
 
-		id = a2dp_source_config(a2dp->session, a2dp_config_complete,
-					client->caps, client);
+		if (!a2dp->sep) {
+			error("seid %d not opened", client->seid);
+			goto failed;
+		}
+
+		id = a2dp_source_config(a2dp->session, a2dp->sep,
+					a2dp_config_complete, client->caps,
+					client);
 		client->cancel = a2dp_source_cancel;
 		break;
 
 	case TYPE_HEADSET:
 		hs = &client->d.hs;
 
-		switch (client->access_mode) {
-		case BT_CAPABILITIES_ACCESS_MODE_READ:
-			hs->lock = HEADSET_LOCK_READ;
-			break;
-		case BT_CAPABILITIES_ACCESS_MODE_WRITE:
-			hs->lock = HEADSET_LOCK_WRITE;
-			break;
-		case BT_CAPABILITIES_ACCESS_MODE_READWRITE:
-			hs->lock = HEADSET_LOCK_READ | HEADSET_LOCK_WRITE;
-			break;
-		default:
-			hs->lock = 0;
-			break;
+		if (!hs->locked) {
+			error("seid %d not opened", client->seid);
+			goto failed;
 		}
 
 		id = headset_config_stream(dev, headset_setup_complete,
-					hs->lock, client);
+						client);
 		client->cancel = headset_cancel_stream;
 		break;
 
@@ -781,7 +947,7 @@ static void start_resume(struct audio_device *dev, struct unix_client *client)
 		}
 
 		if (!a2dp->sep) {
-			error("Unable to get a sep");
+			error("seid not opened");
 			goto failed;
 		}
 
@@ -795,7 +961,7 @@ static void start_resume(struct audio_device *dev, struct unix_client *client)
 		hs = &client->d.hs;
 
 		id = headset_request_stream(dev, headset_resume_complete,
-					hs->lock, client);
+						client);
 		client->cancel = headset_cancel_stream;
 		break;
 
@@ -850,7 +1016,7 @@ static void start_suspend(struct audio_device *dev, struct unix_client *client)
 		hs = &client->d.hs;
 
 		id = headset_suspend_stream(dev, headset_suspend_complete,
-					hs->lock, client);
+						client);
 		client->cancel = headset_cancel_stream;
 		break;
 
@@ -870,30 +1036,98 @@ failed:
 	unix_ipc_error(client, BT_STOP_STREAM, EIO);
 }
 
+static void close_complete(struct audio_device *dev, void *user_data)
+{
+	struct unix_client *client = user_data;
+	char buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_close_rsp *rsp = (void *) buf;
+
+	memset(buf, 0, sizeof(buf));
+
+	rsp->h.type = BT_RESPONSE;
+	rsp->h.name = BT_CLOSE;
+	rsp->h.length = sizeof(*rsp);
+
+	unix_ipc_sendmsg(client, &rsp->h);
+
+	return;
+}
+
+static void start_close(struct audio_device *dev, struct unix_client *client,
+			gboolean reply)
+{
+	struct a2dp_data *a2dp;
+	struct headset_data *hs;
+
+	switch (client->type) {
+	case TYPE_HEADSET:
+		hs = &client->d.hs;
+
+		if (client->dev && hs->locked) {
+			headset_unlock(client->dev, client->lock);
+			hs->locked = FALSE;
+		}
+		break;
+	case TYPE_SOURCE:
+	case TYPE_SINK:
+		a2dp = &client->d.a2dp;
+
+		if (client->cb_id > 0)
+			avdtp_stream_remove_cb(a2dp->session, a2dp->stream,
+								client->cb_id);
+		if (a2dp->sep) {
+			a2dp_sep_unlock(a2dp->sep, a2dp->session);
+			a2dp->sep = NULL;
+		}
+		if (a2dp->session) {
+			avdtp_unref(a2dp->session);
+			a2dp->session = NULL;
+		}
+		break;
+	default:
+		error("No known services for device");
+		goto failed;
+	}
+
+	if (reply)
+		close_complete(dev, client);
+
+	return;
+
+failed:
+	if (reply)
+		unix_ipc_error(client, BT_STOP_STREAM, EINVAL);
+}
+
 static void handle_getcapabilities_req(struct unix_client *client,
 					struct bt_get_capabilities_req *req)
 {
 	struct audio_device *dev;
-	bdaddr_t bdaddr;
+	bdaddr_t src, dst;
+	int err = EIO;
 
-	str2ba(req->device, &bdaddr);
-
-	if (client->interface) {
-		error("Got GET_CAPABILITIES for an initialized client");
+	if (!check_nul(req->source) || !check_nul(req->destination) ||
+			!check_nul(req->object)) {
+		err = EINVAL;
 		goto failed;
 	}
 
+	str2ba(req->source, &src);
+	str2ba(req->destination, &dst);
+
 	if (req->transport == BT_CAPABILITIES_TRANSPORT_SCO)
 		client->interface = g_strdup(AUDIO_HEADSET_INTERFACE);
 	else if (req->transport == BT_CAPABILITIES_TRANSPORT_A2DP)
 		client->interface = g_strdup(AUDIO_SINK_INTERFACE);
 
-	if (!manager_find_device(&bdaddr, NULL, FALSE))
+	if (!manager_find_device(req->object, &src, &dst, NULL, FALSE))
 		goto failed;
 
-	dev = manager_find_device(&bdaddr, client->interface, TRUE);
+	dev = manager_find_device(req->object, &src, &dst, client->interface,
+				TRUE);
 	if (!dev && (req->flags & BT_FLAG_AUTOCONNECT))
-		dev = manager_find_device(&bdaddr, client->interface, FALSE);
+		dev = manager_find_device(req->object, &src, &dst,
+					client->interface, FALSE);
 
 	if (!dev) {
 		error("Unable to find a matching device");
@@ -906,12 +1140,100 @@ static void handle_getcapabilities_req(struct unix_client *client,
 		goto failed;
 	}
 
+	client->seid = req->seid;
+
 	start_discovery(dev, client);
 
 	return;
 
 failed:
-	unix_ipc_error(client, BT_GET_CAPABILITIES, EIO);
+	unix_ipc_error(client, BT_GET_CAPABILITIES, err);
+}
+
+static int handle_sco_open(struct unix_client *client, struct bt_open_req *req)
+{
+	if (!client->interface)
+		client->interface = g_strdup(AUDIO_HEADSET_INTERFACE);
+	else if (!g_str_equal(client->interface, AUDIO_HEADSET_INTERFACE))
+		return -EIO;
+
+	debug("open sco - object=%s source=%s destination=%s lock=%s%s",
+			strcmp(req->object, "") ? req->object : "ANY",
+			strcmp(req->source, "") ? req->source : "ANY",
+			strcmp(req->destination, "") ? req->destination : "ANY",
+			req->lock & BT_READ_LOCK ? "read" : "",
+			req->lock & BT_WRITE_LOCK ? "write" : "");
+
+	return 0;
+}
+
+static int handle_a2dp_open(struct unix_client *client, struct bt_open_req *req)
+{
+	if (!client->interface)
+		client->interface = g_strdup(AUDIO_SINK_INTERFACE);
+	else if (!g_str_equal(client->interface, AUDIO_SINK_INTERFACE))
+		return -EIO;
+
+	debug("open a2dp - object=%s source=%s destination=%s lock=%s%s",
+			strcmp(req->object, "") ? req->object : "ANY",
+			strcmp(req->source, "") ? req->source : "ANY",
+			strcmp(req->destination, "") ? req->destination : "ANY",
+			req->lock & BT_READ_LOCK ? "read" : "",
+			req->lock & BT_WRITE_LOCK ? "write" : "");
+
+	return 0;
+}
+
+static void handle_open_req(struct unix_client *client, struct bt_open_req *req)
+{
+	struct audio_device *dev;
+	bdaddr_t src, dst;
+	int err = 0;
+
+	if (!check_nul(req->source) || !check_nul(req->destination) ||
+			!check_nul(req->object)) {
+		err = EINVAL;
+		goto failed;
+	}
+
+	str2ba(req->source, &src);
+	str2ba(req->destination, &dst);
+
+	if (req->seid > BT_A2DP_SEID_RANGE) {
+		err = handle_sco_open(client, req);
+		if (err < 0) {
+			err = -err;
+			goto failed;
+		}
+	} else {
+		err = handle_a2dp_open(client, req);
+		if (err < 0) {
+			err = -err;
+			goto failed;
+		}
+	}
+
+	if (!manager_find_device(req->object, &src, &dst, NULL, FALSE))
+		goto failed;
+
+	dev = manager_find_device(req->object, &src, &dst, client->interface,
+				TRUE);
+	if (!dev)
+		dev = manager_find_device(req->object, &src, &dst,
+					client->interface, FALSE);
+
+	if (!dev)
+		goto failed;
+
+	client->seid = req->seid;
+	client->lock = req->lock;
+
+	start_open(dev, client);
+
+	return;
+
+failed:
+	unix_ipc_error(client, BT_SET_CONFIGURATION, err ? : EIO);
 }
 
 static int handle_sco_transport(struct unix_client *client,
@@ -922,9 +1244,6 @@ static int handle_sco_transport(struct unix_client *client,
 	else if (!g_str_equal(client->interface, AUDIO_HEADSET_INTERFACE))
 		return -EIO;
 
-	debug("config sco - device = %s access_mode = %u", req->device,
-			req->access_mode);
-
 	return 0;
 }
 
@@ -951,10 +1270,7 @@ static int handle_a2dp_transport(struct unix_client *client,
 
 	client->caps = g_slist_append(client->caps, media_transport);
 
-	debug("config a2dp - device = %s access_mode = %u", req->device,
-			req->access_mode);
-
-	if (req->codec.type == BT_A2DP_CODEC_MPEG12) {
+	if (req->codec.type == BT_A2DP_MPEG12_SINK) {
 		mpeg_capabilities_t *mpeg = (void *) &req->codec;
 
 		memset(&mpeg_cap, 0, sizeof(mpeg_cap));
@@ -971,12 +1287,8 @@ static int handle_a2dp_transport(struct unix_client *client,
 		media_codec = avdtp_service_cap_new(AVDTP_MEDIA_CODEC, &mpeg_cap,
 							sizeof(mpeg_cap));
 
-		debug("codec mpeg12 - frequency = %u channel_mode = %u "
-			"layer = %u crc = %u mpf = %u bitrate = %u",
-			mpeg_cap.frequency, mpeg_cap.channel_mode,
-			mpeg_cap.layer, mpeg_cap.crc, mpeg_cap.mpf,
-			mpeg_cap.bitrate);
-	} else if (req->codec.type == BT_A2DP_CODEC_SBC) {
+		print_mpeg12(&mpeg_cap);
+	} else if (req->codec.type == BT_A2DP_SBC_SINK) {
 		sbc_capabilities_t *sbc = (void *) &req->codec;
 
 		memset(&sbc_cap, 0, sizeof(sbc_cap));
@@ -994,12 +1306,7 @@ static int handle_a2dp_transport(struct unix_client *client,
 		media_codec = avdtp_service_cap_new(AVDTP_MEDIA_CODEC, &sbc_cap,
 							sizeof(sbc_cap));
 
-		debug("codec sbc - frequency = %u channel_mode = %u "
-			"allocation = %u subbands = %u blocks = %u "
-			"bitpool = %u", sbc_cap.frequency,
-			sbc_cap.channel_mode, sbc_cap.allocation_method,
-			sbc_cap.subbands, sbc_cap.block_length,
-			sbc_cap.max_bitpool);
+		print_sbc(&sbc_cap);
 	} else
 		return -EINVAL;
 
@@ -1011,17 +1318,14 @@ static int handle_a2dp_transport(struct unix_client *client,
 static void handle_setconfiguration_req(struct unix_client *client,
 					struct bt_set_configuration_req *req)
 {
-	struct audio_device *dev;
-	bdaddr_t bdaddr;
 	int err = 0;
 
-	if (!req->access_mode) {
-		err = EINVAL;
+	if (req->codec.seid != client->seid) {
+		error("Unable to set configuration: seid %d not opened",
+				client->seid);
 		goto failed;
 	}
 
-	str2ba(req->device, &bdaddr);
-
 	if (req->codec.transport == BT_CAPABILITIES_TRANSPORT_SCO) {
 		err = handle_sco_transport(client, req);
 		if (err < 0) {
@@ -1036,19 +1340,10 @@ static void handle_setconfiguration_req(struct unix_client *client,
 		}
 	}
 
-	if (!manager_find_device(&bdaddr, NULL, FALSE))
+	if (!client->dev)
 		goto failed;
 
-	dev = manager_find_device(&bdaddr, client->interface, TRUE);
-	if (!dev)
-		dev = manager_find_device(&bdaddr, client->interface, FALSE);
-
-	if (!dev)
-		goto failed;
-
-	client->access_mode = req->access_mode;
-
-	start_config(dev, client);
+	start_config(client->dev, client);
 
 	return;
 
@@ -1084,6 +1379,20 @@ failed:
 	unix_ipc_error(client, BT_STOP_STREAM, EIO);
 }
 
+static void handle_close_req(struct unix_client *client,
+				struct bt_close_req *req)
+{
+	if (!client->dev)
+		goto failed;
+
+	start_close(client->dev, client, TRUE);
+
+	return;
+
+failed:
+	unix_ipc_error(client, BT_CLOSE, EIO);
+}
+
 static void handle_control_req(struct unix_client *client,
 					struct bt_control_req *req)
 {
@@ -1105,8 +1414,6 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 	bt_audio_msg_header_t *msghdr = (void *) buf;
 	struct unix_client *client = data;
 	int len;
-	struct a2dp_data *a2dp = &client->d.a2dp;
-	struct headset_data *hs = &client->d.hs;
 	const char *type, *name;
 
 	if (cond & G_IO_NVAL)
@@ -1114,26 +1421,7 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 
 	if (cond & (G_IO_HUP | G_IO_ERR)) {
 		debug("Unix client disconnected (fd=%d)", client->sock);
-		switch (client->type) {
-		case TYPE_HEADSET:
-			if (client->dev && hs->locked) {
-				headset_unlock(client->dev, hs->lock);
-				hs->locked = FALSE;
-			}
-			break;
-		case TYPE_SOURCE:
-		case TYPE_SINK:
-			if (a2dp->sep) {
-				a2dp_sep_unlock(a2dp->sep, a2dp->session);
-				a2dp->sep = NULL;
-			}
-			break;
-		default:
-			break;
-		}
 
-		if (client->cancel && client->req_id > 0)
-			client->cancel(client->dev, client->req_id);
 		goto failed;
 	}
 
@@ -1160,6 +1448,10 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 		handle_getcapabilities_req(client,
 				(struct bt_get_capabilities_req *) msghdr);
 		break;
+	case BT_OPEN:
+		handle_open_req(client,
+				(struct bt_open_req *) msghdr);
+		break;
 	case BT_SET_CONFIGURATION:
 		handle_setconfiguration_req(client,
 				(struct bt_set_configuration_req *) msghdr);
@@ -1172,6 +1464,10 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 		handle_streamstop_req(client,
 				(struct bt_stop_stream_req *) msghdr);
 		break;
+	case BT_CLOSE:
+		handle_close_req(client,
+				(struct bt_close_req *) msghdr);
+		break;
 	case BT_CONTROL:
 		handle_control_req(client,
 				(struct bt_control_req *) msghdr);
@@ -1185,6 +1481,7 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 
 failed:
 	clients = g_slist_remove(clients, client);
+	start_close(client->dev, client, FALSE);
 	client_free(client);
 	return FALSE;
 }
@@ -1274,6 +1571,8 @@ void unix_exit(void)
 {
 	g_slist_foreach(clients, (GFunc) client_free, NULL);
 	g_slist_free(clients);
-	close(unix_sock);
-	unix_sock = -1;
+	if (unix_sock >= 0) {
+		close(unix_sock);
+		unix_sock = -1;
+	}
 }
