@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -47,6 +47,12 @@ static void formfactor_reply(DBusPendingCall *call, void *user_data)
 
 	reply = dbus_pending_call_steal_reply(call);
 
+	if (dbus_set_error_from_message(NULL, reply) == TRUE) {
+		error("Failed to access HAL");
+		dbus_message_unref(reply);
+		return;
+	}
+
 	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_STRING, &formfactor,
 						DBUS_TYPE_INVALID) == FALSE) {
 		error("Wrong formfactor arguments");
@@ -124,6 +130,8 @@ static int hal_probe(struct btd_adapter *adapter)
 
 	dbus_pending_call_set_notify(call, formfactor_reply, adapter, NULL);
 
+	dbus_pending_call_unref(call);
+
 	dbus_message_unref(message);
 
 	return 0;
@@ -150,4 +158,4 @@ static void hal_exit(void)
 	btd_unregister_adapter_driver(&hal_driver);
 }
 
-BLUETOOTH_PLUGIN_DEFINE("hal", hal_init, hal_exit)
+BLUETOOTH_PLUGIN_DEFINE("hal", VERSION, hal_init, hal_exit)
