@@ -123,6 +123,8 @@ struct btd_adapter {
 	gboolean already_up;		/* adapter was already up on init */
 
 	gboolean off_requested;		/* DEVDOWN ioctl was called */
+
+	uint8_t svc_cache;		/* For bluetoothd startup */
 };
 
 static void adapter_set_pairable_timeout(struct btd_adapter *adapter,
@@ -1192,6 +1194,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	char str[249], srcaddr[18];
+	uint32_t class;
 	gboolean value;
 	char **devices;
 	int i;
@@ -1224,6 +1227,12 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &property);
 
+	/* Class */
+	class = adapter->dev.class[0] |
+			adapter->dev.class[1] << 8 |
+			adapter->dev.class[2] << 16;
+	dict_append_entry(&dict, "Class", DBUS_TYPE_UINT32, &class);
+
 	/* Powered */
 	value = adapter->up ? TRUE : FALSE;
 	dict_append_entry(&dict, "Powered", DBUS_TYPE_BOOLEAN, &value);
@@ -1434,9 +1443,6 @@ static DBusMessage *list_devices(DBusConnection *conn,
 	for (l = adapter->devices; l; l = l->next) {
 		struct btd_device *device = l->data;
 
-		if (device_is_temporary(device))
-			continue;
-
 		dev_path = device_get_path(device);
 
 		dbus_message_iter_append_basic(&array_iter,
@@ -1465,8 +1471,8 @@ static DBusMessage *cancel_device_creation(DBusConnection *conn,
 	device = adapter_find_device(adapter, address);
 	if (!device)
 		return g_dbus_create_error(msg,
-				ERROR_INTERFACE ".DoesNotExist",
-				"Device does not exist");
+				ERROR_INTERFACE ".NotInProgress",
+				"Device creation not in progress");
 
 	if (!device_is_temporary(device) || !device_is_bonding(device, sender))
 		return not_authorized(msg);
@@ -2060,6 +2066,8 @@ proceed:
 		load_connections(adapter, dd);
 
 		adapter->initialized = TRUE;
+
+		adapter_update(adapter, 0, FALSE);
 	}
 
 	if (dev_down) {
@@ -2169,10 +2177,15 @@ setup:
 	hci_send_cmd(dd, OGF_LINK_POLICY, OCF_READ_DEFAULT_LINK_POLICY,
 								0, NULL);
 
-	if (hci_test_bit(HCI_INQUIRY, &di.flags))
-		adapter->state |= STD_INQUIRY;
-	else
-		adapter->state &= ~STD_INQUIRY;
+	if (hci_test_bit(HCI_INQUIRY, &di.flags)) {
+		debug("inquiry_cancel at adapter startup");
+		inquiry_cancel(dd, HCI_REQ_TIMEOUT);
+	} else if (!adapter->initialized && adapter->already_up) {
+		debug("periodic_inquiry_exit at adapter startup");
+		periodic_inquiry_exit(dd, HCI_REQ_TIMEOUT);
+	}
+
+	adapter->state &= ~STD_INQUIRY;
 
 	adapter_setup(adapter, dd);
 	err = adapter_up(adapter, dd);
@@ -2291,14 +2304,26 @@ int adapter_stop(struct btd_adapter *adapter)
 	return 0;
 }
 
-int adapter_update(struct btd_adapter *adapter)
+int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
+							gboolean starting)
 {
 	struct hci_dev *dev = &adapter->dev;
 	int dd;
+	uint8_t svclass;
 
 	if (dev->ignore)
 		return 0;
 
+	if (starting || !adapter->initialized) {
+		adapter->svc_cache = new_svc;
+		return 0;
+	}
+
+	if (new_svc)
+		svclass = new_svc;
+	else
+		svclass = adapter->svc_cache;
+
 	dd = hci_open_dev(adapter->dev_id);
 	if (dd < 0) {
 		int err = -errno;
@@ -2307,6 +2332,9 @@ int adapter_update(struct btd_adapter *adapter)
 		return err;
 	}
 
+	if (svclass)
+		set_service_classes(dd, adapter->dev.class, svclass);
+
 	update_ext_inquiry_response(dd, dev);
 
 	hci_close_dev(dd);
@@ -2326,6 +2354,10 @@ int adapter_get_class(struct btd_adapter *adapter, uint8_t *cls)
 int adapter_set_class(struct btd_adapter *adapter, uint8_t *cls)
 {
 	struct hci_dev *dev = &adapter->dev;
+	uint32_t class;
+
+	if (adapter->svc_cache)
+		adapter->svc_cache = 0;
 
 	if (memcmp(dev->class, cls, 3) == 0)
 		return 0;
@@ -2334,6 +2366,11 @@ int adapter_set_class(struct btd_adapter *adapter, uint8_t *cls)
 
 	write_local_class(&adapter->bdaddr, cls);
 
+	class = cls[0] | (cls[1] << 8) | (cls[2] << 16);
+
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Class", DBUS_TYPE_UINT32, &class);
+
 	return 0;
 }
 
@@ -2621,6 +2658,7 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 {
 	const gchar *path = adapter_get_path(adapter);
 	gboolean powered, discoverable, pairable;
+	uint8_t real_class[3];
 	int dd;
 
 	if (adapter->scan_mode == scan_mode)
@@ -2680,10 +2718,14 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 		goto done;
 	}
 
+	memcpy(real_class, adapter->dev.class, 3);
+	if (adapter->svc_cache)
+		real_class[2] = adapter->svc_cache;
+
 	if (discoverable && adapter->pairable)
-		set_limited_discoverable(dd, adapter->dev.class, TRUE);
+		set_limited_discoverable(dd, real_class, TRUE);
 	else if (!discoverable)
-		set_limited_discoverable(dd, adapter->dev.class, FALSE);
+		set_limited_discoverable(dd, real_class, FALSE);
 
 	hci_close_dev(dd);
 done:
