@@ -71,7 +71,8 @@ int adapter_start(struct btd_adapter *adapter);
 
 int adapter_stop(struct btd_adapter *adapter);
 
-int adapter_update(struct btd_adapter *adapter);
+int adapter_update(struct btd_adapter *adapter, uint8_t cls,
+						gboolean starting);
 
 int adapter_get_class(struct btd_adapter *adapter, uint8_t *cls);
 
@@ -139,3 +140,4 @@ const char *adapter_any_get_path(void);
 const char *btd_adapter_any_request_path(void);
 void btd_adapter_any_release_path(void);
 gboolean adapter_is_pairable(struct btd_adapter *adapter);
+gboolean adapter_powering_down(struct btd_adapter *adapter);
